/******************************************************************************

	COPYRIGHT(C) JONAS 'SORTIE' TERMANSEN 2011.

	This file is part of LibMaxsi.

	LibMaxsi is free software: you can redistribute it and/or modify it under
	the terms of the GNU Lesser General Public License as published by the Free
	Software Foundation, either version 3 of the License, or (at your option)
	any later version.

	LibMaxsi is distributed in the hope that it will be useful, but WITHOUT ANY
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
	FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for
	more details.

	You should have received a copy of the GNU Lesser General Public License
	along with LibMaxsi. If not, see <http://www.gnu.org/licenses/>.

	memory.cpp
	Useful functions for manipulating memory, as well as the implementation
	of the heap.

******************************************************************************/

#include "platform.h"
#include "memory.h"

#ifdef SORTIX_KERNEL
#include <sortix/memorymanagement.h>
#endif

namespace Maxsi
{
	namespace Memory
	{
		// This magic word is useful to see if the program has written into the
		// chunk's header or footer, which means the program has malfunctioned
		// and ought to be terminated.
#ifdef PLATFORM_X64
		const size_t Magic = 0xDEADDEADDEADDEADULL;
#else
		const size_t Magic = 0xDEADDEAD;
#endif
	
		// All requsted sizes must be a multiple of this.
		const size_t Alignment = 8ULL;

		// Returns the index of the most significant set bit.
		inline size_t BSR(size_t Value) { size_t Result; asm("bsr %0, %1" : "=r"(Result) : "r"(Value)); return Result; }

		// Returns the index of the least significant set bit.
		inline size_t BSF(size_t Value) { size_t Result; asm("bsf %0, %1" : "=r"(Result) : "r"(Value)); return Result; }

		// NOTE: BSR(N) = Log2RoundedDown(N).

		// NOTE: BSR(N-1)+1 = Log2RoundedUp(N), N > 1.
		// PROOF: If X=2^N is a power of two, then BSR(X) = Log2(X) = N.
		// If X is not a power of two, then BSR(X) will return the previous
		// power of two.
		// If we want the next power of two for a non-power-of-two then
		// BSR(X)+1 would work. However, this causes problems for X=2^N, where
		// it returns a value one too big.
		// BSR(X-1)+1 = BSR(X)+1 give the same value for all X that is not a
		// power of two. For X=2^N, BSR(X-1)+1 will be one lower, which is
		// exactly what we want. QED.

		// Now declare some structures that are put around the allocated data.
		struct UnusedChunkHeader
		{
			size_t Size;
			UnusedChunkHeader* NextUnused;
		};

		struct UnusedChunkFooter
		{
			void* Unused;
			size_t Size;
		};

		struct UsedChunkHeader
		{
			size_t Magic;
			size_t Size;
		};

		struct UsedChunkFooter
		{
			size_t Magic;
			size_t Size;
		};
		
		// How much extra space will our headers use?
		const size_t ChunkOverhead = sizeof(UsedChunkHeader) + sizeof(UsedChunkFooter);
		size_t GetChunkOverhead() { return ChunkOverhead; }

		// A bin for each power of two.
		UnusedChunkHeader* Bins[BITS(size_t)];

		// INVARIANT: A bit is set for each bin if it contains anything.
		size_t BinContainsChunks;

		// And south of everything, there is nothing allocated, at which
		// extra space can be added on-demand.
		const byte* Wilderness;
		size_t WildernessSize;

		// Initializes the heap system.
		void Init()
		{
			Memory::Set(Bins, 0, sizeof(Bins));
			Wilderness = NULL;
			WildernessSize = 0;
			BinContainsChunks = 0;
		}

		// Expands the wilderness block (the heap grows downwards).
		bool ExpandWilderness(size_t NeededSize)
		{
			if ( NeededSize <= WildernessSize ) { return true; }

#ifdef SORTIX_KERNEL
			// TODO: Allocate system pages here! (TODO: Implement the required system calls).
#else

#endif
			return false;
		}

		// Allocates a continious memory region of Size bytes that must be deallocated using Free.
		DUAL_FUNCTION(void*, malloc, Allocate, (size_t Size))
		{
			// Always allocate a multiple of alignment (round up to nearest aligned size).
			// INVARIANT: Size is always aligned after this statement.
			Size = (Size + Alignment - 1) & ~(Alignment);

			// Account for the overhead of the headers and footers.
			Size += ChunkOverhead;

			// Find the index of the smallest usable bin.
			// INVARIANT: Size is always at least ChunkOverhead bytes, thus
			// Size-1 will never be 0, and thus BSR is always defined.
			size_t MinBinIndex = BSR(Size-1)+1;

			// Make a bitmask that filters away all bins that are too small.
			size_t MinBinMask = ~((1 << MinBinIndex) - 1);
			
			// Now filter all bins away that are too small.
			size_t AvailableBins = BinContainsChunks & MinBinMask;

			// Does any bin contain an usable chunk?
			if ( AvailableBins > 0 )
			{
				// Now find the smallest usable bin.
				size_t BinIndex = BSF( AvailableBins );		

				// And pick the first thing in it.
				UnusedChunkHeader* Chunk = Bins[BinIndex];

				// Increment the bin's linked list.
				Bins[BinIndex] = Chunk->NextUnused;

				// Find the size of this bin (also the value of this bins flag).
				size_t BinSize = 1 << BinIndex;

				// If we emptied the bin, then remove the flag that says this bin is usable.
				if ( Chunk->NextUnused == NULL ) { BinContainsChunks ^= BinSize; }

				// Figure out where the chunk ends.
				uintptr_t ChunkEnd = ((uintptr_t) Chunk) + Chunk->Size;

				// If we were to split the chunk into one used part, and one unused part, then
				// figure out where the unused part would begin.
				uintptr_t NewChunkStart = ((uintptr_t) Chunk) + Size;

				// INVARIANT: NewChunkStart is always less or equal to ChunkEnd,
				// because Size is less or equal to Chunk->Size (otherwise this
				// chunk could not have been selected above).
				// INVARIANT: NewChunkStart is aligned because Chunk is (this
				// is made sure when chunks are made when expanding into the
				// wilderness, when splitting blocks, and when combining them),
				// and because Size is aligned (first thing we make sure).

				// Find the size of the possible new chunk.
				size_t NewChunkSize = ChunkEnd - NewChunkStart;

				UsedChunkHeader* ResultHeader = (UsedChunkHeader*) Chunk;

				// See if it's worth it to split the chunk into two, if any space is left in it.
				if ( ChunkEnd - NewChunkStart >= sizeof(UnusedChunkHeader) + sizeof(UnusedChunkFooter) )
				{
					// Figure out which bin to put the new chunk in.
					size_t NewBinIndex = BSR(NewChunkSize);

					// Mark that a chunk is available for this bin.
					BinContainsChunks |= (1 << NewBinIndex);

					// Now write some headers and footers for the new chunk.
					UnusedChunkHeader* NewChunkHeader = (UnusedChunkHeader*) NewChunkStart;
					NewChunkHeader->Size = NewChunkSize;
					NewChunkHeader->NextUnused = Bins[NewBinIndex];
					UnusedChunkFooter* NewChunkFooter = (UnusedChunkFooter*) (ChunkEnd - sizeof(UnusedChunkFooter));
					NewChunkFooter->Size = NewChunkSize;
					NewChunkFooter->Unused = NULL;

					// Put the new chunk in front of our linked list.
					Bins[NewBinIndex] = NewChunkHeader;

					// We need to modify our resulting chunk to be smaller.
					ResultHeader->Size = Size;
				}

				// Set the required magic values.
				UsedChunkFooter* ResultFooter = (UsedChunkFooter*) ( ((byte*) ResultHeader) + ResultHeader->Size - sizeof(UsedChunkFooter) );
				ResultHeader->Magic = Magic;
				ResultFooter->Magic = Magic;
				ResultFooter->Size = Size;

				return ((byte*) ResultHeader) + sizeof(UnusedChunkHeader);
			}
			else
			{
				// We have no free chunks that are big enough, let's expand our heap into the unknown, if possible.
				if ( WildernessSize < Size && !ExpandWilderness(Size) ) { return NULL; }

				// Write some headers and footers around our newly allocated data.
				UsedChunkHeader* ResultHeader = (UsedChunkHeader*) (Wilderness + WildernessSize - Size);
				UsedChunkFooter* ResultFooter = (UsedChunkFooter*) (Wilderness + WildernessSize - sizeof(UsedChunkFooter));
				WildernessSize -= Size;

				ResultHeader->Size = Size;
				ResultHeader->Magic = Magic;
				ResultFooter->Size = Size;
				ResultFooter->Magic = Magic;

				return ((byte*) ResultHeader) + sizeof(UsedChunkHeader);
			}	
		}

		// Frees a continious memory region allocated by Allocate.
		DUAL_FUNCTION(void, free, Free, (void* Addr))
		{
			// Just ignore NULL-pointers.
			if ( Addr == NULL ) { return; }

			// Restore the chunk information structures.
			UsedChunkHeader* ChunkHeader = (UsedChunkHeader*) ( ((byte*) Addr) - sizeof(UsedChunkHeader) );
			UsedChunkFooter* ChunkFooter = (UsedChunkFooter*) ( ((byte*) ChunkHeader) + ChunkHeader->Size - sizeof(UsedChunkFooter) );
		
			// If you suspect a chunk of bein' a witch, report them immediately. I cannot stress that enough. Witchcraft will not be tolerated.
			if ( ChunkHeader->Magic != Magic || ChunkFooter->Magic != Magic || ChunkHeader->Size != ChunkFooter->Size )
			{
				// TODO: Report witchcraft (terminating the process is probably a good idea).
			}

			// TODO: Combine this chunk with its neighbors, if they are also unused.

			// Calculate which bin this chunk belongs to.
			size_t BinIndex = BSR(ChunkHeader->Size);

			// Mark that a chunk is available for this bin.
			BinContainsChunks |= (1 << BinIndex);

			// Now put this chunk back in the linked list.
			ChunkFooter->Magic = NULL;
			ChunkHeader->Magic = (size_t) Bins[BinIndex];
			Bins[BinIndex] = (UnusedChunkHeader*) ChunkHeader;
		}

		DUAL_FUNCTION(void*, memcpy, Copy, (void* Dest, void* Src, size_t Length))
		{
			char* D = (char*) Dest; char* S = (char*) Src;

			for ( size_t I = 0; I < Length; I++ )
			{
				D[I] = S[I];
			}

			return Dest;
		}

		DUAL_FUNCTION(void*, memset, Set, (void* Dest, int Value, size_t Length))
		{
			byte* D = (byte*) Dest;

			for ( size_t I = 0; I < Length; I++ )
			{
				D[I] = Value & 0xFF;
			}

			return Dest;
		}
	}
}
