/******************************************************************************

	COPYRIGHT(C) JONAS 'SORTIE' TERMANSEN 2011.

	This file is part of Sortix.

	Sortix is free software: you can redistribute it and/or modify it under the
	terms of the GNU General Public License as published by the Free Software
	Foundation, either version 3 of the License, or (at your option) any later
	version.

	Sortix is distributed in the hope that it will be useful, but WITHOUT ANY
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
	FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
	details.

	You should have received a copy of the GNU General Public License along
	with Sortix. If not, see <http://www.gnu.org/licenses/>.

	kernel.cpp
	A common interface shared by all devices that can be printed text to.

******************************************************************************/

#include "platform.h"
#include <libmaxsi/memory.h>
#include <libmaxsi/string.h>
#include "globals.h"
#include "iprintable.h"
#include "log.h"
#include "panic.h"
#include "descriptor_tables.h"
#include "iirqhandler.h"
#include "time.h"
#include "keyboard.h"
#include "multiboot.h"
#include "memorymanagement.h"
#include "scheduler.h"
#include "syscall.h"
#include "pong.h"
#include "pci.h"
#include "uart.h"

void* RunApplication(void* Parameter);

namespace Sortix
{
	void KernelConstructors()
	{
		GLog = new (GLog) Log;
		GPanic = new (GPanic) PanicClass;
	}

	extern "C" void KernelInit(unsigned long Magic, multiboot_info_t* BootInfo)
	{
#ifdef JSSORTIX
		// TODO: Make JSVM multiboot compliant.
		multiboot_info_t MBInfo; BootInfo = &MBInfo;
		multiboot_memory_map_t MBMMap;

		MBMMap.addr = 0x100000;
		MBMMap.len = 0xC00000;
		MBMMap.type = MULTIBOOT_MEMORY_AVAILABLE;
		MBMMap.size = sizeof(MBMMap) - sizeof(MBMMap.size);
		BootInfo->flags = MULTIBOOT_INFO_MEM_MAP;
		BootInfo->mmap_addr = (multiboot_uint32_t) &MBMMap;
		BootInfo->mmap_length = sizeof(MBMMap);

		// Initialize the serial driver.
		UART::Init();

#define BSOD
#ifdef BSOD

		UART::WriteChar(27);
		UART::WriteChar(91);
		UART::WriteChar(48 + 4);
		UART::WriteChar(48 + 4);
		UART::WriteChar(109);

		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("Windows Boot Manager has experienced a problem.                                 ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("    Status: 0xc000000f                                                          ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("    Info: An error occured during transferring execution.                       ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("You can try to recover the system with the Microsoft Windows System Recoverry   ");
		GLog->Print("Tools. (You might need to restart the system manually).                         ");
		GLog->Print("                                                                                ");
		GLog->Print("If the problem continious, please contact your system administrator or computer ");
		GLog->Print("Manufacturer.                                                                   ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");

		JSSortix::Exit();

		return;

#endif

		// Just a test to see if the color system works! Make the BG red!
		UART::WriteChar(27);
		UART::WriteChar(91);
		UART::WriteChar(48 + 4);
		UART::WriteChar(48 + 1);
		UART::WriteChar(109);

#endif

		// Initialize the global object's by calling their constructors.
		KernelConstructors();

		// Initialize the kernel log.
		GLog->Init();

		GLog->Print("                                                       _                        ");
		GLog->Print("                                                      / \\                       ");
		GLog->Print("                  /\\    /\\                           /   \\                      ");
		GLog->Print("                 /  \\  /  \\                          |   |                      ");
		GLog->Print("                /    \\/    \\                         |   |                      ");
		GLog->Print("               |  O    O    \\_______________________ /   |                      ");
		GLog->Print("               |                                         |                      ");
		GLog->Print("               | \\_______/                               /                      ");
		GLog->Print("                \\                                       /                       ");
		GLog->Print("                  ------       ---------------      ---/                        ");
		GLog->Print("                       /       \\             /      \\                           ");
		GLog->Print("                      /         \\           /        \\                          ");
		GLog->Print("                     /           \\         /          \\                         ");
		GLog->Print("                    /_____________\\       /____________\\                        ");
		GLog->Print("                                                                                ");
#ifdef JSSORTIX
		GLog->Print("                     = THE SORTIX KERNEL - PONG EDITION =                       ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
		GLog->Print("                                                                                ");
#endif

		// Initialize the GDT and TSS structures.
		GDT::Init();

		// Initialize the interrupt descriptor tables.
		IDT::Init();

		if ( BootInfo == NULL ) { Panic("kernel.cpp: The bootinfo structure was NULL. Are your bootloader multiboot compliant?"); }

		// Initialize the paging.
		Page::Init(BootInfo);

#ifndef JSSORTIX
		PCI::Init();
#endif

		// Initialize virtual memory. TODO: This is not fully working yet.
		//VirtualMemory::Init();

#ifdef PONG
		Pong::Init();
		Keyboard::Init();		
		Time::Init();
		while ( true ) { }
#endif

		// Initialize the keyboard.
		Keyboard::Init();

		// Initialize system calls.
		Syscall::Init();

		// Initialize the scheduler.
		Scheduler::Init();

		if ( Scheduler::CreateThread(NULL, RunApplication) == NULL )
		{
			Panic("Could not create a sample thread!");
		}

		// Lastly set up the timer driver and we are ready to run the OS.
		Time::Init();

		// Run the OS.
		Scheduler::MainLoop();
	}
}
