/******************************************************************************

	COPYRIGHT(C) JONAS 'SORTIE' TERMANSEN 2011.

	This file is part of Sortix.

	Sortix is free software: you can redistribute it and/or modify it under the
	terms of the GNU General Public License as published by the Free Software
	Foundation, either version 3 of the License, or (at your option) any later
	version.

	Sortix is distributed in the hope that it will be useful, but WITHOUT ANY
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
	FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
	details.

	You should have received a copy of the GNU General Public License along
	with Sortix. If not, see <http://www.gnu.org/licenses/>.

	scheduler.h
	Handles the creation and management of threads.

******************************************************************************/

#include "platform.h"
#include <libmaxsi/memory.h>
#include "panic.h"
#include "scheduler.h"
#include "globals.h"
#include "multiboot.h"
#include "memorymanagement.h"
#include "descriptor_tables.h"

#include "iprintable.h"
#include "log.h"

#ifndef PLATFORM_X86
#error This CPU has no scheduler available
#endif

namespace Sortix
{
	namespace Scheduler
	{
		// This is a very small thread that does absoluting nothing!
		char NoopThreadData[sizeof(ThreadInfo)];
		ThreadInfo* NoopThread;
		const size_t NoopThreadStackLength = 8;
		size_t NoopThreadStack[NoopThreadStackLength];
		void NoopFunction() { while ( true ) { } }

		// Linked lists that implements a very simple scheduler.
		ThreadInfo* CurrentThread;
		ThreadInfo* FirstRunnableThread;
		ThreadInfo* FirstUnrunnableThread;
		ThreadInfo* FirstSleeping;
		size_t AllocatedThreadId;

		// Initializes the scheduling subsystem.
		void Init()
		{
			CurrentThread = NULL;
			FirstRunnableThread = NULL;
			FirstUnrunnableThread = NULL;
			FirstSleeping = NULL;
			AllocatedThreadId = 1;

			// Initialize the thread that does nothing.
			NoopThread = new ((void*) NoopThreadData) ThreadInfo(NULL, 0, NULL, 0);
			NoopThread->SetState(NOOP);
#ifdef PLATFORM_X86
			NoopThread->Registers.useresp = (uint32_t) NoopThreadStack + NoopThreadStackLength;
			NoopThread->Registers.ebp = (uint32_t) NoopThreadStack + NoopThreadStackLength;
			NoopThread->Registers.eip = (uint32_t) &NoopFunction; // NoopFunction;
#endif

			// Allocate and set up a stack for the kernel to use during interrupts.
			void* KernelStackPage = Page::Get();
			if ( KernelStackPage == NULL ) { Panic("scheduler.cpp: could not allocate kernel interrupt stack for tss!"); }
			size_t* KernelStack = ((size_t*) KernelStackPage) + 4096 / sizeof(size_t);
			GDT::SetKernelStack(KernelStack);
		}

		// Once the init process is spawned and IRQ0 is enabled, this process
		// simply awaits an IRQ0 and then we shall be scheduling.
		void MainLoop()
		{
			// Simply wait for the next IRQ0 and then the OS will run.
			while ( true ) { }
		}

		ThreadInfo* CreateThread(ProcessInfo* Process, ThreadEntry Start, void* Parameter1, void* Parameter2, size_t StackSize)
		{
			// The current default stack size is 4096 bytes.
			if ( StackSize == SIZE_MAX ) { StackSize = 4096; }

			// TODO: We only support stacks of up to one page!
			if ( 4096 < StackSize ) { StackSize = 4096; }
			
			// TODO: Use the proper memory management systems using new and delete instead of these hacks!
			// TODO: These allocations might NOT be thread safe!
			void* ThreadPage = Page::Get();
			if ( ThreadPage == NULL ) { return NULL; }
			void* StackPage = Page::Get();
			if ( StackPage == NULL ) { Page::Put(ThreadPage); return NULL; }

			// Allocate a stack for this thread.
			size_t* Stack = (size_t*) StackPage;
			size_t StackLength = StackSize / sizeof(size_t);

			// Create a new thread data structure.
			ThreadInfo* Thread = new (ThreadPage) ThreadInfo(Process, AllocatedThreadId++, Stack, StackLength);

#ifdef PLATFORM_X86
			// Prepare the parameters for the entry function (C calling convention).
			//Stack[StackLength - 1] = (size_t) 0xFACE; // Parameter2
			Stack[StackLength - 1] = (size_t) Parameter2; // Parameter2
			Stack[StackLength - 2] = (size_t) Parameter1; // Parameter1
			Stack[StackLength - 3] = (size_t) Thread->GetId(); // This thread's id.
			Stack[StackLength - 4] = (size_t) 0xB00C; // Eip
			Thread->Registers.ebp = Thread->Registers.useresp = (uint32_t) (Stack + StackLength - 4); // Point to the last word used on the stack.
			Thread->Registers.eip = (uint32_t) Start; // Point to our entry function.
			//Thread->Registers.edi = 0x42;
			//Thread->Registers.eax = 0x69;
			//Thread->Registers.ebx = 0x69;
#endif

			// Mark the thread as running, which adds it to the scheduler's linked list.
			Thread->SetState(RUNNABLE);

			return Thread;
		}

		ThreadInfo* PopNextThread()
		{
			//GLog->PrintF("PopNextThread(): CurrentThread = 0x%p, FirstRunnableThread = 0x%p, FirstRunnableThread->PrevThread = 0x%p, FirstRunnableThread->NextThread = 0x%p\n", CurrentThread, FirstRunnableThread, FirstRunnableThread->PrevThread, FirstRunnableThread->NextThread);

			if ( FirstRunnableThread == NULL )
			{
				return NoopThread;
			}
			else if ( CurrentThread == FirstRunnableThread )
			{
				FirstRunnableThread = FirstRunnableThread->NextThread;
			}

			return FirstRunnableThread;
		}

		void Switch(CPU::InterruptRegisters* R, uintmax_t TimePassed)
		{
			//GLog->PrintF("Scheduling while at eip=0x%p...", R->eip);

			WakeSleeping(TimePassed);

			// Find the next thread to be run.
			ThreadInfo* NextThread = PopNextThread();

			//if ( NextThread == NoopThread ) { PanicF("Going to NoopThread! Noop=0x%p, First=0x%p -> 0x%p, Unrun=0x%p", NoopThread, FirstRunnableThread, FirstRunnableThread->NextThread, FirstUnrunnableThread); }

			// If the next thread happens to be the current one, simply do nothing.
			if ( CurrentThread != NextThread )
			{
				// Save the hardware registers of the current thread.
				if ( CurrentThread != NULL )
				{
					//GLog->PrintF("Switching from thread at 0x%p to thread at 0x%p\n", CurrentThread, NextThread);
					CurrentThread->SaveRegisters(R);
				}

				// TODO: If applicable, switch the virtual address space.
				CurrentThread = NextThread;

				// Load the hardware registers of the next thread.
				//GLog->PrintF("Switching to thread at 0x%p\n", NextThread);
				NextThread->LoadRegisters(R);
			}

			uint32_t RPL = ( CurrentThread != NoopThread ) ? 0x3 : 0x0;

			RPL = 0x3;

			// Jump into user-space!
			R->ds = 0x20 | RPL; // Set the data segment and Requested Privilege Level.
			R->cs = 0x18 | RPL; // Set the code segment and Requested Privilege Level.
			R->ss = 0x20 | RPL; // Set the stack segment and Requested Privilege Level.
			R->eflags |= 0x200; // Enable the enable interrupts flag in EFLAGS!

			//GLog->PrintF("ds=0x%x, edi=0x%x, esi=0x%x, ebp=0x%x, esp=0x%x, ebx=0x%x, edx=0x%x, ecx=0x%x, eax=0x%x, int_no=0x%x, err_code=0x%x, eip=0x%x, cs=0x%x, eflags=0x%x, useresp=0x%x, ss=0x%x\n", R->ds, R->edi, R->esi, R->ebp, R->esp, R->ebx, R->edx, R->ecx, R->eax, R->int_no, R->err_code, R->eip, R->cs, R->eflags, R->useresp, R->ss);

			size_t* Stack = (size_t*) R->useresp;

/*
			// TODO: HACK: Currently ESP is not properly set after we return
			// from the interrupt. So we call a helper function that restores
			// it by storing it in EAX, then restoring EAX, and restoring EIP,
			// then we should have returned successfully.
			
			Stack[-1] = (size_t) &PrintRegistersAndDie; // R->eip;
			Stack[-2] = (size_t) R->eax;
			R->eax = (uint32_t) (Stack - 2);
			R->eip = (uint32_t) &RestoreStack;

			GLog->PrintF("ds=0x%x, edi=0x%x, esi=0x%x, ebp=0x%x, esp=0x%x, ebx=0x%x, edx=0x%x, ecx=0x%x, eax=0x%x, int_no=0x%x, err_code=0x%x, eip=0x%x, cs=0x%x, eflags=0x%x, useresp=0x%x, ss=0x%x\n", R->ds, R->edi, R->esi, R->ebp, R->esp, R->ebx, R->edx, R->ecx, R->eax, R->int_no, R->err_code, R->eip, R->cs, R->eflags, R->useresp, R->ss);
*/

			//GLog->PrintF("Stack = {0x%p, 0x%p, 0x%p, 0x%p, 0x%p, 0x%p, 0x%p, 0x%p, 0x%p, 0x%p, 0x%p, 0x%p, 0x%p, 0x%p, 0x%p, 0x%p}\n", Stack[0], Stack[1], Stack[2], Stack[3], Stack[4], Stack[5], Stack[6], Stack[7], Stack[8], Stack[9], Stack[10], Stack[11], Stack[12], Stack[13], Stack[14], Stack[15]);

			//GLog->PrintF(" resuming at eip=0x%p\n", R->eip);
		}

		void WakeSleeping(uintmax_t TimePassed)
		{
			while ( FirstSleeping != NULL )
			{		
				if ( TimePassed < FirstSleeping->SleepMilisecondsLeft ) { FirstSleeping->SleepMilisecondsLeft -= TimePassed; break; }

				TimePassed -= FirstSleeping->SleepMilisecondsLeft;
				FirstSleeping->SleepMilisecondsLeft = 0;
				FirstSleeping->SetState(RUNNABLE);
				ThreadInfo* Next = FirstSleeping->NextSleepingThread;
				FirstSleeping->NextSleepingThread = NULL;
				FirstSleeping = Next;					
			}
		}

		void ExitThread(ThreadInfo* Thread, void* Result)
		{
			// TODO: What do we do with the result parameter?
			Thread->~ThreadInfo();
			Page::Put(Thread);

			if ( Thread == CurrentThread ) { CurrentThread = NULL; }
		}

		ThreadInfo* GetCurrentThread()
		{
			return CurrentThread;
		}

		ProcessInfo* GetCurrentProcess()
		{
			if ( CurrentThread != NULL ) { return CurrentThread->GetProcess(); } else { return NULL; }
		}

		void SysCreateThread(CPU::InterruptRegisters* R)
		{
#ifdef PLATFORM_X86
			ThreadInfo* Thread = CreateThread(GetCurrentProcess(), (ThreadEntry) R->ebx, (void*) R->ecx, (void*) R->edx, (size_t) R->edi);
			R->eax = (Thread != NULL) ? Thread->GetId() : 0;
#endif
		}

		void SysExitThread(CPU::InterruptRegisters* R)
		{
#ifdef PLATFORM_X86
			ExitThread(GetCurrentThread(), (void*) R->ebx);
			Switch(R, 0);
#endif
		}

		void SysSleep(CPU::InterruptRegisters* R)
		{
			//GLog->PrintF("<SysSleep>\n");
#ifdef PLATFORM_X86
			intmax_t TimeToSleep = ((uintmax_t) R->ebx) * 1000ULL;
#endif
			if ( TimeToSleep == 0 ) { return; }
			GetCurrentThread()->Sleep(TimeToSleep);
			Switch(R, 0);
			//GLog->PrintF("</SysSleep>\n");
		}

		void SysUSleep(CPU::InterruptRegisters* R)
		{
#ifdef PLATFORM_X86
			intmax_t TimeToSleep = ((uintmax_t) R->ebx) / 1000ULL;
#endif
			if ( TimeToSleep == 0 ) { return; }
			GetCurrentThread()->Sleep(TimeToSleep);
			Switch(R, 0);
		}
	}

	ThreadInfo::ThreadInfo(ProcessInfo* NewProcess, size_t NewId, size_t* NewStack, size_t NewStackLength)
	{
		Process = NewProcess;
		Id = NewId;
		Stack = NewStack;
		StackLength = NewStackLength;
		State = INFANT;
		PrevThread = this;
		NextThread = this;
		InThisList = NULL;
	}

	ThreadInfo::~ThreadInfo()
	{
		if ( InThisList != NULL && *InThisList == this )
		{
			*InThisList = ( NextThread != this ) ? NextThread : NULL;
			InThisList = NULL;
		}
		if ( NextThread != this ) { PrevThread->NextThread = NextThread; NextThread->PrevThread = PrevThread; PrevThread = this; NextThread = this; }

		Page::Put(Stack);
	}
	
	void ThreadInfo::Relink(ThreadInfo** First)
	{
		if ( InThisList != NULL && *InThisList == this )
		{
			*InThisList = ( NextThread != this ) ? NextThread : NULL;
			InThisList = NULL;
		}
		if ( NextThread != this ) { PrevThread->NextThread = NextThread; NextThread->PrevThread = PrevThread; PrevThread = this; NextThread = this; }

		//GLog->PrintF("Relinking 0x%p to 0x%p (0x%p)\n", this, First, *First);
	
		InThisList = First;	
		if ( *First != NULL )
		{
			(*First)->PrevThread->NextThread = this;
			PrevThread = (*First)->PrevThread;
			(*First)->PrevThread = this;
			NextThread = *First;
		}
		else
		{
			*First = this;
		}
	}

	void ThreadInfo::SetState(ThreadState NewState)
	{
		if ( NewState == State ) { return; }

		if ( NewState == RUNNABLE )
		{
			Relink(&Scheduler::FirstRunnableThread);
		}
		else if ( ( NewState == UNRUNNABLE ) /*&& ( State != UNRUNNABLE || State != WAITING )*/ )
		{
			Relink(&Scheduler::FirstUnrunnableThread);
		}

		State = NewState;
	}

	void ThreadInfo::SaveRegisters(CPU::InterruptRegisters* Src)
	{
		Registers.eip = Src->eip; Registers.useresp = Src->useresp; Registers.eax = Src->eax; Registers.ebx = Src->ebx; Registers.ecx = Src->ecx; Registers.edx = Src->edx; Registers.edi = Src->edi; Registers.esi = Src->esi; Registers.ebp = Src->ebp;
	}

	void ThreadInfo::LoadRegisters(CPU::InterruptRegisters* Dest)
	{
		Dest->eip = Registers.eip; Dest->useresp = Registers.useresp; Dest->eax = Registers.eax; Dest->ebx = Registers.ebx; Dest->ecx = Registers.ecx; Dest->edx = Registers.edx; Dest->edi = Registers.edi; Dest->esi = Registers.esi; Dest->ebp = Registers.ebp;
	}

	void ThreadInfo::Sleep(uintmax_t Miliseconds)
	{
		if ( Miliseconds == 0 ) { return; }

		ThreadInfo* Thread = Scheduler::FirstSleeping;

		if ( Thread == NULL )
		{
			Scheduler::FirstSleeping = this;
		}
		else if ( Miliseconds < Thread->SleepMilisecondsLeft )
		{
			Scheduler::FirstSleeping = this;
			NextSleepingThread = Thread;
			Thread->SleepMilisecondsLeft -= Miliseconds;
		}
		else
		{
			while ( true )
			{
				if ( Thread->NextSleepingThread == NULL )
				{
					Thread->NextSleepingThread = this; break;
				}
				else if ( Miliseconds < Thread->NextSleepingThread->SleepMilisecondsLeft )
				{
					Thread->NextSleepingThread->SleepMilisecondsLeft -= Miliseconds;
					NextSleepingThread = Thread->NextSleepingThread;
					Thread->NextSleepingThread = this; break;
				}
				else
				{
					Miliseconds -= Thread->SleepMilisecondsLeft;
					Thread = Thread->NextSleepingThread;
				}
			}
		}

		SleepMilisecondsLeft = Miliseconds;
		SetState(UNRUNNABLE);	
	}
}

