/******************************************************************************

	COPYRIGHT(C) JONAS 'SORTIE' TERMANSEN 2011.

	This file is part of Sortix.

	Sortix is free software: you can redistribute it and/or modify it under the
	terms of the GNU General Public License as published by the Free Software
	Foundation, either version 3 of the License, or (at your option) any later
	version.

	Sortix is distributed in the hope that it will be useful, but WITHOUT ANY
	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
	FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
	details.

	You should have received a copy of the GNU General Public License along
	with Sortix. If not, see <http://www.gnu.org/licenses/>.

	uart.h
	A simple serial terminal driver.

******************************************************************************/

#include "platform.h"
#include "uart.h"

namespace Sortix
{
	namespace UART
	{
		const nat TXR = 0; // Transmit register
		const nat RXR = 0; // Receive register
		const nat IER = 1; // Interrupt Enable 
		const nat IIR = 2; // Interrupt ID
		const nat FCR = 2; // FIFO control
		const nat LCR = 3; // Line control
		const nat MCR = 4; // Modem control
		const nat LSR = 5; // Line Status
		const nat MSR = 6; // Modem Status
		const nat DLL = 0; // Divisor Latch Low 
		const nat DLM = 1; // Divisor latch High

		const nat LCR_DLAB = 0x80; // Divisor latch access bit
		const nat LCR_SBC = 0x40; // Set break control
		const nat LCR_SPAR = 0x20; // Stick parity (?)
		const nat LCR_EPAR = 0x10; // Even parity select
		const nat LCR_PARITY = 0x08; // Parity Enable
		const nat LCR_STOP = 0x04; // Stop bits: 0=1 bit, 1=2 bits
		const nat LCR_WLEN5 = 0x00; // Wordlength: 5 bits
		const nat LCR_WLEN6 = 0x01; // Wordlength: 6 bits
		const nat LCR_WLEN7 = 0x02; // Wordlength: 7 bits
		const nat LCR_WLEN8 = 0x03; // Wordlength: 8 bits

		const nat LSR_TEMT = 0x40; // Transmitter empty
		const nat LSR_THRE = 0x20; // Transmit-hold-register empty
		const nat LSR_READY = 0x1;

		const nat Port = 0x3f8;

		const nat BASE_BAUD = 1843200/16;
		const nat BOTH_EMPTY = LSR_TEMT | LSR_THRE;

#ifdef JSSORTIX
		uint16_t VGALastFrame[80 * 25];
#endif

		nat ProbeBaud(nat Port)
		{
			uint8_t lcr = X86::InPortB(Port + LCR);
			X86::OutPortB(Port + LCR, lcr | LCR_DLAB);
			uint8_t dll = X86::InPortB(Port + DLL);
			uint8_t dlm = X86::InPortB(Port + DLM);
			X86::OutPortB(Port + LCR, lcr);
			nat quot = (dlm << 8) | dll;

			return BASE_BAUD / quot;
		}

		void WaitForEmptyBuffers(nat Port)
		{
			while ( true )
			{
				nat Status = X86::InPortB(Port + LSR);

				if ( (Status & BOTH_EMPTY) == BOTH_EMPTY )
				{
					return;
				}
			}
		}

		nat Baud;

		void Init()
		{
#ifdef JSSORTIX
			for ( nat I = 0; I < 80 * 25; I++ ) { VGALastFrame[I] = 0; }
#endif

			// This crashes the JS VM, so don't do it.
			return;

			Baud = ProbeBaud(Port);

			X86::OutPortB(Port + LCR, 0x3); // 8n1
			X86::OutPortB(Port + IER, 0); // No interrupt
			X86::OutPortB(Port + FCR, 0); // No FIFO
			X86::OutPortB(Port + MCR, 0x3); // DTR + RTS

			nat Divisor = 115200 / Baud;
			uint8_t C = X86::InPortB(Port + LCR);
			X86::OutPortB(Port + LCR, C | LCR_DLAB);
			X86::OutPortB(Port + DLL, Divisor & 0xFF);
			X86::OutPortB(Port + DLM, (Divisor >> 8) & 0xFF);
			X86::OutPortB(Port + LCR, C & ~LCR_DLAB);
		}

		void Read(uint8_t* Buffer, size_t Size)
		{
			// Save the IER and disable interrupts.
			nat ier = X86::InPortB(Port + IER);
			X86::OutPortB(Port + IER, 0);

			for ( size_t I = 0; I < Size; I++ )
			{
				while ( ! ( X86::InPortB(Port + LSR) & LSR_READY ) ) { }

				Buffer[I] = X86::InPortB(Port);
			}

			// Wait for transmitter to become empty and restore the IER.
			WaitForEmptyBuffers(Port);
			X86::OutPortB(Port + IER, ier);
		}

		void Write(const uint8_t* Buffer, size_t Size)
		{
			// Save the IER and disable interrupts.
			nat ier = X86::InPortB(Port + IER);
			X86::OutPortB(Port + IER, 0);

			for ( size_t I = 0; I < Size; I++ )
			{
				WaitForEmptyBuffers(Port);

				X86::OutPortB(Port, Buffer[I]);
			}

			// Wait for transmitter to become empty and restore the IER.
			WaitForEmptyBuffers(Port);
			X86::OutPortB(Port + IER, ier);
		}

		void WriteChar(char C)
		{
			// Save the IER and disable interrupts.
			nat ier = X86::InPortB(Port + IER);
			X86::OutPortB(Port + IER, 0);

			WaitForEmptyBuffers(Port);

			X86::OutPortB(Port, C);

			// Wait for transmitter to become empty and restore the IER.
			WaitForEmptyBuffers(Port);
			X86::OutPortB(Port + IER, ier);
		}

		int TryPopChar()
		{
			// Save the IER and disable interrupts.
			nat ier = X86::InPortB(Port + IER);
			X86::OutPortB(Port + IER, 0);

			int Result = -1;
		
			if ( X86::InPortB(Port + LSR) & LSR_READY )
			{
				Result = X86::InPortB(Port);
			}

			// Wait for transmitter to become empty and restore the IER.
			WaitForEmptyBuffers(Port);
			X86::OutPortB(Port + IER, ier);

			return Result;
		}

#ifdef JSSORTIX
		void RenderVGA()
		{
			// Set the cursor to (0,0)
			UART::WriteChar(27);
			UART::WriteChar(91);
			UART::WriteChar(72);

			uint16_t* Source = (uint16_t*) 0xB8000;

			// Change from VGA color to another color system.
			nat ConversionTable[16] = { 0, 4, 2, 6, 1, 5, 3, 7, 0, 4, 2, 6, 1, 5, 3, 7 };

			nat LastColor = 1337;

			bool Skipped = false;

			for ( nat Y = 0; Y < 25; Y++)
			{
				for ( nat X = 0; X < 80; X++ )
				{
					nat Element = Source[Y * 80 + X];
					nat OldElement = VGALastFrame[Y * 80 + X];
					VGALastFrame[Y * 80 + X] = Element;

					if ( Element == OldElement ) { Skipped = true; continue; }

					nat NewColor = ConversionTable[ (Element >> 12) & 0xF ] << 3 | ConversionTable[ (Element >> 8) & 0xF ];

					// Update the position if we skipped some characters.
					if ( Skipped )
					{
						UART::WriteChar(27);
						UART::WriteChar(91);
						UART::WriteChar(48 + (Y+1) / 10);
						UART::WriteChar(48 + (Y+1) % 10);
						UART::WriteChar(59);
						UART::WriteChar(48 + (X+1) / 10);
						UART::WriteChar(48 + (X+1) % 10);
						UART::WriteChar(72);
						Skipped = false;					
					}

					// Change the color if we need to.
					if ( LastColor != NewColor )
					{
						UART::WriteChar(27);
						UART::WriteChar(91);
						UART::WriteChar(48 + 3);
						UART::WriteChar(48 + NewColor % 8 );
						UART::WriteChar(59);
						UART::WriteChar(48 + 4);
						UART::WriteChar(48 + NewColor / 8 );
						UART::WriteChar(109);

						LastColor = NewColor;
					}

					Element &= 0x7F;

					// Filter away any not printable characters.
					if ( Element < 32 || Element > 126 ) { Element = '?'; }

					UART::WriteChar(Element & 0x7F);
				}
			}
		}
#endif
	}
}
